@@ -2,7 +2,7 @@ package Mail::Milter::Authentication::Handler::DMARC;
 use strict;
 use warnings;
 use base 'Mail::Milter::Authentication::Handler';
-our $VERSION = '2.20180511'; # VERSION
+our $VERSION = '2.20180521'; # VERSION
 
 use Data::Dumper;
 use English qw{ -no_match_vars };
@@ -20,6 +20,7 @@ my $PSL_CHECKED_TIME;
 sub default_config {
     return {
         'hide_none'      => 0,
+        'use_arc'        => 1,
         'hard_reject'    => 0,
         'no_list_reject' => 1,
         'whitelisted'    => [],
@@ -182,15 +183,56 @@ sub _process_dmarc_for {
         return;
     }
 
+    my $have_arc = ( $self->is_handler_loaded( 'ARC' ) );
+    if ( $have_arc ) {
+        # Does our ARC handler have the necessary methods?
+        $have_arc = 0 unless $self->get_handler('ARC')->can( 'get_trusted_arc_authentication_results' );
+    }
+    my $used_arc = 0;
+    $have_arc = 0 if ! $config->{ 'use_arc' };
+
     # Add the SPF Results Object
     eval {
         my $spf = $self->get_handler('SPF');
         if ( $spf ) {
-            $dmarc->spf(
-                'domain' => $spf->{'dmarc_domain'},
-                'scope'  => $spf->{'dmarc_scope'},
-                'result' => $spf->{'dmarc_result'},
-            );
+
+            if ( ! $have_arc ) {
+                $dmarc->spf(
+                    'domain' => $spf->{'dmarc_domain'},
+                    'scope'  => $spf->{'dmarc_scope'},
+                    'result' => $spf->{'dmarc_result'},
+                );
+            }
+            else {
+                if ( $spf->{'dmarc_result'} eq 'pass' && lc $spf->{'dmarc_domain'} eq lc $header_domain ) {
+                    # Have a matching local entry, use it.
+                    ## TODO take org domains into consideration here
+                    $dmarc->spf(
+                        'domain' => $spf->{'dmarc_domain'},
+                        'scope'  => $spf->{'dmarc_scope'},
+                        'result' => $spf->{'dmarc_result'},
+                    );
+                }
+                elsif ( my $arc_spf = $self->get_handler('ARC')->get_trusted_spf_results() ) {
+                    # Pull from ARC if we can
+                    $used_arc = 1;
+                    push @$arc_spf, {
+                        'domain' => $spf->{'dmarc_domain'},
+                        'scope'  => $spf->{'dmarc_scope'},
+                        'result' => $spf->{'dmarc_result'},
+                    };
+                    $dmarc->spf( $arc_spf );
+                }
+                else {
+                    # Nothing else matched, use the local entry anyway
+                    $dmarc->spf(
+                        'domain' => $spf->{'dmarc_domain'},
+                        'scope'  => $spf->{'dmarc_scope'},
+                        'result' => $spf->{'dmarc_result'},
+                    );
+                }
+            }
+
         }
         else {
             $dmarc->{'spf'} = [];
@@ -204,21 +246,40 @@ sub _process_dmarc_for {
 
     # Add the DKIM Results
     my $dkim_handler = $self->get_handler('DKIM');
-    if ( $dkim_handler->{'failmode'} ) {
-        $dmarc->{'dkim'} = [];
+    if ( $have_arc ) {
+        my @dkim_values;
+        my $arc_values = $self->get_handler('ARC')->get_trusted_dkim_results();
+        if ( $arc_values ) {
+            foreach my $arc_value ( @$arc_values ) {
+                push @dkim_values, $arc_value;
+            }
+        }
+        $dmarc->{'dkim'} = \@dkim_values;
+        # Add the local DKIM object is it exists
+        if ( $dkim_handler->{'has_dkim'} ) {
+            my $dkim_object = $self->get_object('dkim');
+            if ( $dkim_object ) {
+                $dmarc->dkim( $dkim_object );
+            }
+        }
     }
-    elsif ( $dkim_handler->{'has_dkim'} ) {
-        my $dkim_object = $self->get_object('dkim');
-        if ( $dkim_object ) {
-            $dmarc->dkim( $dkim_object );
+    else {
+        if ( $dkim_handler->{'failmode'} ) {
+            $dmarc->{'dkim'} = [];
+        }
+        elsif ( $dkim_handler->{'has_dkim'} ) {
+            my $dkim_object = $self->get_object('dkim');
+            if ( $dkim_object ) {
+                $dmarc->dkim( $dkim_object );
+            }
+            else {
+                $dmarc->{'dkim'} = [];
+            }
         }
         else {
             $dmarc->{'dkim'} = [];
         }
     }
-    else {
-        $dmarc->{'dkim'} = [];
-    }
 
     # Run the Validator
     my $dmarc_result = $dmarc->validate();
@@ -246,6 +307,11 @@ sub _process_dmarc_for {
     $dmarc_policy = 'none' if ! $dmarc_policy;
     $self->dbgout( 'DMARCPolicy', $dmarc_policy, LOG_INFO );
 
+    ## TODO decide if this is the right thing to do
+    if ( $used_arc ) {
+        $dmarc_result->reason( 'type' => 'trusted_forwarder', 'comment' => 'Trusted ARC chain considered' );
+    }
+
     my $policy_override;
 
     # Reject mail and/or set policy override reasons
@@ -296,11 +362,15 @@ sub _process_dmarc_for {
         #if ( $policy_override ) {
         #    push @comments, $self->format_header_entry( 'override', $policy_override );
         #}
+        if ( $used_arc ) {
+            push @comments, 'Trusted ARC chain considered';
+        }
 
         if ( @comments ) {
             $header->add_child( Mail::AuthenticationResults::Header::Comment->new()->safe_set_value( join( ',', @comments ) ) );
         }
 
+
         $header->add_child( Mail::AuthenticationResults::Header::SubEntry->new()->set_key( 'header.from' )->safe_set_value( $header_domain ) );
         $self->_add_dmarc_header( $header );
     }
@@ -312,6 +382,7 @@ sub _process_dmarc_for {
         'policy'         => $dmarc_policy,
         'is_list'        => ( $self->{'is_list'}      ? '1' : '0' ),
         'is_whitelisted' => ( $self->is_whitelisted() ? '1' : '0'),
+        'used_arc'       => ( $used_arc ? '1' : '0' ),
     };
     $self->metric_count( 'dmarc_total', $metric_data );
 
@@ -483,6 +554,11 @@ sub header_callback {
 sub eom_requires {
     my ($self) = @_;
     my @requires = qw{ DKIM };
+
+    if ( $self->is_handler_loaded( 'ARC' ) ) {
+        push @requires, 'ARC';
+    }
+
     return \@requires;
 }
 
@@ -533,6 +609,7 @@ sub eom_callback {
                     $self->_add_dmarc_header( $header );
                 }
             }
+            $self->check_timeout();
         }
     }
 
@@ -637,7 +714,7 @@ Mail::Milter::Authentication::Handler::DMARC
 
 =head1 VERSION
 
-version 2.20180511
+version 2.20180521
 
 =head1 DESCRIPTION
 
@@ -656,6 +733,7 @@ This handler requires the SPF and DKIM handlers to be installed and active.
                 "dkim:bad.forwarder.com",                | (valid) DKIM signing domains can also be whitelisted by
                 "20.30.40.0/24"                          | having an entry such as "dkim:domain.com"
             ],
+            "use_arc"             : 1,                   | Use trusted ARC results if available
             "hide_none"           : 0,                   | Hide auth line if the result is 'none'
             "detect_list_id"      : "1",                 | Detect a list ID and modify the DMARC authentication header
                                                          | to note this, useful when making rules for junking email
